package ZipTie::Adapters::Brocade::vRouter;

use strict;
use warnings;

use ZipTie::Adapters::Brocade::vRouter::AutoLogin;
use ZipTie::Adapters::Brocade::vRouter::Disconnect qw(disconnect);
use ZipTie::Adapters::Brocade::vRouter::Parsers qw(parse_interfaces print_interfaces create_config parse_chassis parse_filters parse_system parse_static_routes parse_object_groups print_object_groups);
use ZipTie::Adapters::Brocade::vRouter::Utils qw(clean_responses);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::ConnectionPath;
use ZipTie::ConnectionPath::Protocol;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;
use ZipTie::Logger;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
use ZipTie::Adapters::BaseAdapter;
our @ISA = qw(ZipTie::Adapters::BaseAdapter);

sub backup
{
	my $package_name = shift;

	# Retrieve the operation XML document that contains all of the IP, protocol, credential, and file server information
	# that is needed to successfully backup a device.
	my $backup_doc = shift;
	my $responses  = {};

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ($connection_path) = ZipTie::Typer::translate_document( $backup_doc, 'connectionPath' );

	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the device prompt that is returned as a result of a successful connection.
	my ( $cli_protocol, $device_prompt_regex ) = _connect($connection_path);

	# Store the found device prompt regular expression as "prompt" on the ZipTie::CLIProtocol object.
	$cli_protocol->set_prompt_by_name( "prompt", $device_prompt_regex );

	# being creating the model output
	#
	# Grab an output filehandle for the model
	my $filehandle = get_model_filehandle( "vRouter", $connection_path->get_ip_address() );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, "common" );
	$printer->open_model();

	# Gather inputs for the core model
	$responses->{showVer}  = _send( $cli_protocol, "show version",   $device_prompt_regex );
	$responses->{showName} = _send( $cli_protocol, "show host name", $device_prompt_regex );

	# Gather chassis details
	$responses->{showMem} = _send( $cli_protocol, "show system memory", $device_prompt_regex );

	# Add the text configuration to the model output
	$responses->{config} = _send( $cli_protocol, "show configuration all | no-more", $device_prompt_regex );

	# Clean control character sequences from the response
	$responses = clean_responses($responses);

	parse_system( $responses, $printer );
	delete( $responses->{showName} );
	parse_chassis( $responses, $printer );
	delete( $responses->{showMem} );

	# clean up the config (strip the prompt and opening command)
	$responses->{config} =~ s/^.*?show configuration.+?[\n\r]//ms;
	$responses->{config} =~ s/$device_prompt_regex//;

	create_config( $responses, $printer );
	parse_object_groups( $responses, $printer );
	print_object_groups( $responses, $printer );
	parse_filters( $responses, $printer );
	delete( $responses->{showVer} );

	$responses->{interfaces} = _send( $cli_protocol, "show interfaces | no-more", $device_prompt_regex );
	parse_interfaces( $responses, $printer );
	print_interfaces( $responses, $printer );
	delete( $responses->{interfaces} );
	
	parse_static_routes($responses,$printer);
	delete( $responses->{config} );

	# Disconnect from the CLI of the specified device
	disconnect($cli_protocol);

	# write the closing tag of the model
	$printer->close_model();

	# Close the model output file handle
	close_model_filehandle($filehandle);
}

sub _connect {
	# Grab our arguments
	my $connection_path = shift;

	# Create a new CLI protocol object
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Make a connection to and successfully authenticate with the vRouter device
	my $device_prompt_regex = ZipTie::Adapters::Brocade::vRouter::AutoLogin::execute( $cli_protocol, $connection_path );

	# Return the created ZipTie::CLIProtocol object and the device prompt encountered after successfully connecting to a device.
	return ( $cli_protocol, $device_prompt_regex );
}

sub _send {
	# Brocade always echos the system prompt initially, and it can screw up our waitfors
	# causing us to bail out too early.  This helper method builds a more strict regex so
	# we never match the first prompt.
	my ( $cliProtocol, $command, $prompt ) = @_;
#	return $cliProtocol->send_and_wait_for( $command, '(?s)' . quotemeta($command) . '.*' . $prompt );
	return $cliProtocol->send_and_wait_for( $command, $prompt );
}

1;
