package ZipTie::Adapters::CheckPoint::CPSMS::Parsers;

use strict;
use warnings;

use Net::CIDR::Set; # for 'group with exclusion' objects
use MIME::Base64 'encode_base64';
use XML::Simple;
use XML::LibXML;
use v5.10;
use ZipTie::Adapters::CheckPoint::CPSMS::CPSMSConverter;
use ZipTie::Adapters::CheckPoint::CPSMS::FileUtilities qw(getFileToString getFilePath);
use ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser;
use ZipTie::Adapters::CheckPoint::OPSEC::Parsers qw(parse_rules_file generate_filter_list get_any);
use ZipTie::Adapters::CheckPoint::Utilities::FileUtility qw(getFileandParse);
use ZipTie::Adapters::Utils qw(create_network_element get_port_expression mask_to_bits get_any_address_address);

use Exporter 'import';
our @EXPORT_OK =
  qw(parse_chassis parse_CPSMS_interfaces parse_CPSMS_object_groups parse_filters parse_system parse_static_routes parse_snmp);

my $LOGGER = ZipTie::Logger::get_logger();
use constant FAUX_SINGLE_INTERFACE_NAME => "eth0";

sub parse_system
{
	my ( $in, $out ) = @_;
	
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	# extract system from network_objects.xml
	$in->{xsl} = "system";
	$in->{device_system} = _cp_to_sed_xml($in, {"ipaddr"=>"ipaddr"});
	
	# get and store CP fw_policies.xml
	$in->{xml} = "fw_policies";  
	$in->{fw_policies} = _cp_to_doc($in);
		
	# get device's ip address 
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# create a new parser for network objects 
	my $parserno = XML::LibXML->new();	
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");
	
	# create a new parser for firewall policies
	my $parserfw = XML::LibXML->new();
	# create a new XPath context and register namespace for firewall policies
	my $XPCFW = _xpcregistration($parserfw,$in->{fw_policies},"fw_policies");
	
	# get gateway's host name, interfaces
	my $gateway = _get_gateway($XPCNO, $ip_addr);
	
	# create a new DOM
	my $dom = $parserno->parse_string($in->{device_system});
	# and set document's root
	my $doc = $dom->documentElement();
	
	# check NAT existence
	_checkNAT($in,$XPCNO, $XPCFW, $gateway);
	
	 # print nodes and elements	
	my @systemNameNodes = $doc->findnodes('//core:systemName');
	if( @systemNameNodes ) {
		my $systemNameNode = $systemNameNodes[-1];
		my $ref = $systemNameNode->toString();
		$out->print_string($ref);
	}
	
	$out->open_element('core:osInfo');
		_printelement($doc,'core:make',$out);
		_printelement($doc,'core:name',$out);
		_printelement($doc,'core:version',$out);
		_printelement($doc,'core:osType',$out);
		if ($in->{isNAT}) {
			# get an addressTranslationParser object
			my $addressTranslationParser = ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser->new ($in);
			my $packetProcessingOrdering = $addressTranslationParser->getPacketProcessingOrdering();
			$out->print_string($packetProcessingOrdering);
		} 
	$out->close_element('core:osInfo');
	_printelement($doc,'core:biosVersion',$out);
	_printelement($doc,'core:deviceType',$out);
}

sub parse_chassis
{
	my ( $in, $out ) = @_;
	
	my $chassis = {};
	# get and store CP statuses.xml
	$in->{xml} = "statuses";  
	$in->{statuses} = _cp_to_doc($in);
	
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	
	# get device's ip address 
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# create a new parser 
	my $parserno = XML::LibXML->new();
	
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");

	# get gateway's host name, interfaces
	my $gateway = _get_gateway($XPCNO, $ip_addr);
	_getClusteringInfo($XPCNO, $gateway);
	
	# create a new parser 
	my $parserst = XML::LibXML->new();
	
	# create a new XPath context and register namespace for statuses
	my $XPCST = _xpcregistration($parserst,$in->{statuses},"statuses");

	# add total memory info to gateway
	_getMemoryStat($XPCST,$gateway);
	my $model = _getModel($XPCNO,$XPCST, $gateway);
	
	my $memory = {
		kind => "RAM",
	};
	
	foreach my $totalmem (@{$gateway->{MemoryStat}}){
		if (exists $totalmem->{totalVirtMem}){
			$memory->{size} = $totalmem->{totalVirtMem};
			push (@{$chassis->{memory}}, $memory);	
		}
	} # end foreach totalmem
	
	# set vendor and model
	$chassis->{'core:asset'}->{'core:assetType'} = 'Chassis';
	$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} = 'Check Point';
	if (defined $model){
		$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $model;
	} else {
		$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = "Unknown";
	}
	
	$out->print_element('chassis', $chassis) if (defined $chassis) ;
		
}

sub parse_CPSMS_object_groups 
{	
	my ( $in, $out ) = @_;
	
	# get device's ip address
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# extract networkGroup from network_objects.xml
	$in->{xml} = "network_objects";
	$in->{xsl} = "networkGroup";
	$in->{networkGroup} = _cp_to_sed_xml($in);
	# get and store CP network_objects.xml
	$in->{network_objects} = _cp_to_doc($in);
	
	# create a new parser for network object
	my $parserno = XML::LibXML->new();
	
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");
		
	# extract pseudo protocolGroup and serviceGroup from services.xml
	$in->{xml} = "services";
	$in->{xsl} = "pseudoProtocolServiceGroup";
	$in->{xmlstring} = _cp_to_sed_xml($in);

	# get and store CP services.xml
	$in->{services} = _cp_to_doc($in);
	
	# create a new parser for services
	my $parsersrv = XML::LibXML->new();
	
	# create a new XPath context and register namespace for services
	my $XPCSRV = _xpcregistration($parsersrv,$in->{services},"services");
	
	# extract protocolGroup and serviceGroup from $in->{xmlstring} xmlstring
	# by default $in->{xml} = "";
	$in->{xsl} = "protocolServiceGroup";
	$in->{protocolServiceGroup} = _cp_to_sed_xmlstring($in);
												
	# create a new parser 
	my $parser = XML::LibXML->new();
	
	# networkGroup
	# create a new XPath context and register namespace
	my $XPC = _xpcregistration($parser, $in->{networkGroup},"ngns");
	
	$out->open_element('objectGroups');
	parse_network_groups ($out, $XPC, $ip_addr, $in, $XPCNO);
		 	
	# create a new XPath context and register namespace
	$XPC = _xpcregistration($parser, $in->{protocolServiceGroup},"psgns");
	
	# protocolGroup
	parse_protocol_groups ($out, $XPC);
	
	# serviceGroup	
	parse_service_groups ($out, $XPCSRV, $in);
		
	$out->close_element('objectGroups');
}

sub parse_filters {
	
	my ($in, $out) = @_;
	
	$out->open_element("filterLists");

	# check NAT existence 
	if ($in->{isNAT}){
		# print out packetFiltering
		# this happens before translating and routing regardless of 
		# 'translate destination on client side' (nat_dst_client_side) switch value from properties.xml
		$out->open_element('filterListsSection');	
			_packetFiltering($in,$out);
		$out->close_element('filterListsSection');
		# print out address-translation
		_addressTranslation ($in, $out);
	} else {
		_packetFiltering($in,$out);
	}
	
	$out->close_element("filterLists");
	
}	

sub _packetFiltering {
	my ( $in, $out ) = @_;
	
	# get and store CP fw_policies.xml
	$in->{xml} = "fw_policies";  
	$in->{fw_policies} = _cp_to_doc($in);
	
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	
	# get and store CP properties.xml
	$in->{xml} = "properties";
	$in->{properties} = _cp_to_doc($in);
	
	# get and store CP services.xml
	$in->{xml} = "services";
	$in->{services} = _cp_to_doc($in);
	
	# get and store CP statuses.xml
	$in->{xml} = "statuses";
	$in->{statuses} = _cp_to_doc($in);
	
	# get device's ip address
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $managementServerIp = $connectionPath->get_management_server_ip_address();
	my $ip_addr = $connectionPath->get_ip_address();
	
	# get CPT directory
	my $cpt_dir = $in->{cpsms_ctx}->{cpt_dir};
	
	# set CPT xml output directory
	my $xml_dir = $in->{cpsms_ctx}->{xml_dir};
	$xml_dir = $xml_dir."/".$managementServerIp . "/". $ip_addr;
	
	# get recording's flag
	my $env_rec = $in->{cpsms_ctx}->{env_rec};
	
	# create parsers, XPath contexts, and register nampespaces
	
	# create a new parser for firewall policies
	my $parserfw = XML::LibXML->new();
	# create a new XPath context and register namespace for firewall policies
	my $XPCFW = _xpcregistration($parserfw,$in->{fw_policies},"fw_policies"); 

	# create a new parser for network object
	my $parserno = XML::LibXML->new();
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");	

	# create a new parser for properties
	my $parserpr = XML::LibXML->new();
	# create a new XPath context and register namespace for properties
	my $XPCPR = _xpcregistration($parserpr,$in->{properties},"properties"); 
	
	# create a new parser for services
	my $parsersrv = XML::LibXML->new();
	# create a new XPath context and register namespace for properties
	my $XPCSRV = _xpcregistration($parsersrv,$in->{services},"services");
	
	# create a new parser for statuses
	my $parserStatuses = XML::LibXML->new();
	# create a new XPath context and register namespace for properties
	my $XPCSTS = _xpcregistration($parserStatuses,$in->{statuses},"statuses");
	
	# get gateway's host name, interfaces
	my $gateway = _get_gateway($XPCNO, $ip_addr);
	
	# get clustering info if exists and store for NAT
	my $clusterInfo = _getClusteringInfo($XPCNO, $gateway);
	$in->{clusterInfo} = $clusterInfo;
	
	# get properties enablement by parsing properties and store the enabled properties elements plus the positional elements
	$in->{enabled_services} = _getPropertiesEnablement($XPCPR);
	
	# initialize default $filterList
	my $filterList = {
        mode => 'stateful',
    };
	
	# generate the appliesTo element that contains the interfaces where the rules are applied
	# +++wayne I don't think this is right.  Aren't filters enabled by applied interface?
	# +++wayne but here I go piling on even if it is a bug.  There's a possibility that we
	# didn't get interface detail for the local system.  In this case, there is a fake
	# name we've concocted for a single syntheized interface
	
	if (defined $gateway->{intfs})	{
		foreach my $intf (@{$gateway->{intfs}}) {
			push(@{$filterList->{appliesTo}}, {source => $intf->{officialname}});
			
		}
	}
	else	{
		push(@{$filterList->{appliesTo}}, {source => FAUX_SINGLE_INTERFACE_NAME});
	}
	my @appliesTo = @{$filterList->{appliesTo}};
	
	my $serviceGroups = $in->{parsed_services};
	
	# if NAT configuration is enabled then set processingPhase
	if ($in->{isNAT}) {
		$in->{NAT}->{settings}->{processingPhase} = "packet-filtering";
	}
	
	# generate OPSEC user rules filterList
	my $processOrder = 0;
	$out->open_element('filterList');
	$out->print_element('defaultAction', 'none');
	$out->print_element('description', 'OPSEC user rules');
	$out->print_element('appliesTo',\@appliesTo);
	
	my $installedPolicy;
	# get installed policy from statuses
	if ($clusterInfo->{isGatewayCluster}){
		foreach my $member (@{$clusterInfo->{cluster_members}}){
			$installedPolicy = _getInstalledPolicy($XPCSTS, $member);
			last if $installedPolicy ne "";
		}
	} else {
		$installedPolicy = _getInstalledPolicy($XPCSTS, $gateway->{hostname});
	}
	# store for NAT
	$in->{installedPolicy} = $installedPolicy;
	# find object nodes from rules and create filterEntry
	RULE: foreach my $libxmlobject ($XPCFW->findnodes("/fw_policies/object[collection='referenced_object: $installedPolicy']/rule/rule/object")){
		my $filterEntry = {};
		# check if rule is disabled
		my $isDisabled = $libxmlobject->findvalue('./disabled');
		if ($isDisabled eq "True"){
			next RULE;
		}
		
		# find object_uid -> ruleUID essential for rule counting
		my $rule_uid = $libxmlobject->findvalue('./object_uid');
		if ($rule_uid ne "" && $rule_uid ne "empty_object_uid") {
			$filterEntry->{ztSignature} = $rule_uid;
		}
		
		my $installObjValue = $libxmlobject->findvalue('./install/object/empty_field_name/empty_field_name');
		my @targetArray = split ("referenced_object: ",$installObjValue);

		# check if is cluster member and compare target with cluster gateways
		my $isGatewayClusterTarget;
		if ($clusterInfo->{isClusterMember} && scalar @{$clusterInfo->{gateway_clusters}}){
			foreach my $gatewayCluster (@{$clusterInfo->{gateway_clusters}}){
				if ($gatewayCluster ~~ @targetArray){
					$isGatewayClusterTarget = 1;
					last;
				}
			}
		}
		if ( ('Any' ~~ @targetArray) ||  ('Gateways' ~~ @targetArray) || ($gateway->{hostname} ~~  @targetArray) ||$isGatewayClusterTarget ) {
		#if ( (@targetArray eq 'Any') ||  ($target eq 'Gateways') || ( $target eq $gateway->{hostname}) ||$isGatewayClusterTarget ) {
			# find rule's destination value
			my $dstValue = $libxmlobject->findvalue('./dst/object/empty_field_name/empty_field_name');
			my @dstArray = split ("referenced_object: ",$dstValue);
			
			foreach my $dst (@dstArray){
				if ($dst eq "Any") {
	    			push (@{$filterEntry->{destinationIpAddr}}, get_any());
				} elsif ($dst ne ""){
					if (defined $in->{networkGroups}->{$dst}){
						push (@{$filterEntry->{destinationIpAddr}}, _set_objectGroupReference($dst));
					} else{
						$LOGGER->debug("_packetFiltering; Object group reference $dst NOT FOUND. Dropping current rule.");
						next RULE;
					}
				};
			}; #end foreach dst
			
			# find rule's services 
			foreach my $libxmlempty_field_name ($libxmlobject->findnodes('./services/object/empty_field_name')) {
				my $empty_field_name = $libxmlempty_field_name->findvalue('./empty_field_name');
				my @services = 	split ("referenced_object: ",$empty_field_name);
				foreach my $service (@services){
					if ($service ne "" && $service !~ /Any/) {
						# check for __src serviceGroups
						_checkServiceGroup($in,$service,$filterEntry);	
					}	
				} # end foreach services
			} # end foreach empty_field_name
			
			# find rule's track method
			my $track = _substring_after($libxmlobject->findvalue('./track/track'),"referenced_object: ");
			if( $track eq "Log"){
				$filterEntry->{log} = "true";
			}else {
				$filterEntry->{log} = "false";
			}

			my $name = $libxmlobject->findvalue('./name');
			if ($name ne "") {
				$filterEntry->{name} = $name;
			}
			
			# find type (permission) values
			my $type = $libxmlobject->findvalue('./action/action/object/type');
			if ($type eq "accept"){
				$type = "permit";
			}
			if ($type eq "reject"){
				$type = "deny";
			}
			# other possible action type drop, encrypt, decrypt, key install
			if ($type ne ""){
				$filterEntry->{permitAction} = $type;
			}

			# find source value
			my $srcValue = $libxmlobject->findvalue('./src/object/empty_field_name/empty_field_name');
			my @srcArray = split ("referenced_object: ",$srcValue);
			
			foreach my $src (@srcArray){
				if ($src eq "Any") {
	    			push (@{$filterEntry->{sourceIpAddr}}, get_any());
				} elsif ($src ne ""){
					if (defined $in->{networkGroups}->{$src}){
						push (@{$filterEntry->{sourceIpAddr}}, _set_objectGroupReference($src));
					} else{
						$LOGGER->debug("_packetFiltering; Object group reference $src NOT FOUND. Dropping current rule.");
						next RULE;
					}
				};	
			}; #end foreach src 
			
		_printFilterEntry($filterEntry, ++$processOrder, $out);
		} # end if target		
	} #foreach libxmlobject
	
	# close the one-time elements for OPSEC user rules	
	$out->print_element('mode', 'stateful');
	$out->print_element('name', 'OPSEC');
	if ($in->{isNAT}) {
		# get an addressTranslationParser object	
		my $addressTranslationParser = ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser->new ($in);
		my $packetProcessingPhase = $addressTranslationParser->getPacketProcessingPhase();
		$out->print_string($packetProcessingPhase);
	} 
	$out->close_element('filterList');
}

sub _getClusteringInfo{
	my ($XPCNO, $gateway) = @_;
	
	my $hostname = $gateway->{hostname};
	my @noObjectArray = $XPCNO->findnodes("/network_objects/object[object_name='$hostname']");
	if (scalar @noObjectArray eq 0){
		return;
	}
	my $noObject = $noObjectArray[0];
	my $type = $noObject->findvalue('./type');
	if ($type eq "gateway_cluster"){
		$gateway->{isGatewayCluster} = 1;
		# get cluster members cluster_members
		my @clusterMembers = $noObject->findnodes('./cluster_members');
		foreach my $member (@clusterMembers){
			my $memberValue = $member->findvalue('.');
			$memberValue = _substring_after($memberValue,"referenced_object: ");
			push @{$gateway->{cluster_members}}, $memberValue;
		}
	} elsif ($type eq "cluster_member"){
		$gateway->{isClusterMember} = 1;
		my @gatewayClusterArray = $XPCNO->findnodes("/network_objects/object[type='gateway_cluster']");
		foreach my $gatewayCluster (@gatewayClusterArray){
			my $gatewayClusterName = $gatewayCluster->findvalue('./object_name');
			# get cluster members cluster_members
			my @clusterMembers = $gatewayCluster->findnodes('./cluster_members');
			foreach my $member (@clusterMembers){
				my $memberValue = $member->findvalue('.');
				$memberValue = _substring_after($memberValue,"referenced_object: ");
				if ($memberValue eq $hostname){
					push @{$gateway->{gateway_clusters}}, $gatewayClusterName;
				}
			}
		}
	}
	return $gateway;
}

sub _getInstalledPolicy {
	my ($XPCSTS, $hostname) = @_;
	
	my $installedPolicy = "";
	foreach my $libxmlobject ($XPCSTS->findnodes('/statuses/object')){
		my $appObj = $libxmlobject->findvalue('./application_obj');
		$appObj= _substring_after($appObj,"referenced_object: ");
		# find the matching hostname cpshared_application
		if ($appObj =~ /firewall_application_$hostname/){
			$installedPolicy = $libxmlobject->findvalue('./fwStatus/object/fwPolicyStat/object/fwPolicyName');
			return $installedPolicy;
		}
	}
}

sub _printFilterEntry {
	my ($filterEntry, $processOrder, $out) = @_;
	$out->open_element ('filterEntry');
	foreach my $val (@{$filterEntry->{destinationIpAddr}}){
		$out->print_element ('destinationIpAddr',$val);
	}
	if ((defined $filterEntry->{destinationService}) and (@{$filterEntry->{destinationService}}) ){
			foreach my $service (@{$filterEntry->{destinationService}}){
				$out->open_element('destinationService');
				$out->print_element('objectGroupReference', $service);
				$out->close_element('destinationService');	
			}		
	}
	$out->print_element('log', $filterEntry->{log});
	$out->print_element('name',$filterEntry->{name}) if defined $filterEntry->{name};
	$out->print_element('primaryAction',$filterEntry->{permitAction});
	$out->print_element ('processOrder', $processOrder);
	if( defined $filterEntry->{protocol} ) {
		my @protocol_names = keys %{ $filterEntry->{protocol} };
		$out->print_element('protocol', \@protocol_names);
	}
	foreach my $val (@{$filterEntry->{sourceIpAddr}}){
		$out->print_element ('sourceIpAddr',$val);
	}
	if ((defined $filterEntry->{sourceService})  and (@{$filterEntry->{sourceService}})){
			foreach my $service (@{$filterEntry->{sourceService}}){
					$out->open_element('sourceService');
					$out->print_element('objectGroupReference', $service);
					$out->close_element('sourceService');	
			}
	} 
	$out->print_element('ztSignature', $filterEntry->{ztSignature}) if defined $filterEntry->{ztSignature};	
	$out->close_element ('filterEntry');
	return;
}

sub _addressTranslation {
	
	my ($in, $out) = @_;

	# set manual NAT as undefined by default
	$in->{manualNAT}->{available} = 0;
			
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	
	# get and store CP properties.xml
	$in->{xml} = "properties";
	$in->{properties} = _cp_to_doc($in); 
	
	# get and store CP fw_policies.xml
	$in->{xml} = "fw_policies";
	$in->{fw_policies} = _cp_to_doc($in); 

	# create a new parser for firewall policies
	my $parserfw = XML::LibXML->new();
	# create a new XPath context and register namespace for firewall policies
	my $XPCFW = _xpcregistration($parserfw,$in->{fw_policies},"fw_policies");
	
	# create a new parser for network object
	my $parserno = XML::LibXML->new();
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");	

	# create a new parser for properties
	my $parserpr = XML::LibXML->new();
	# create a new XPath context and register namespace for properties
	my $XPCPR = _xpcregistration($parserpr,$in->{properties},"properties");

	$in->{isCPSMS} = 1;
	$in->{XPCFW} = $XPCFW;
	$in->{XPCNO} = $XPCNO;
	$in->{XPCPR} = $XPCPR;
	
	# get an addressTranslationParser object
	my $addressTranslationParser = ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser->new ($in);
	# store global NAT settings under $in->{NAT}->{settings}
	$addressTranslationParser->getNATProperties();
	
	#set processingPhase
	$in->{NAT}->{settings}->{processingPhase} = "address-translation";
	my $packetProcessingPhase = $addressTranslationParser->getPacketProcessingPhase();
	
	# check where the automatic NAT is applied, at the server side or at the client side
	if ( defined($in->{NAT}->{settings}->{nat_dst_client_side}) and  $in->{NAT}->{settings}->{nat_dst_client_side} eq 'True')  {
		$out->open_element('filterListsSection');
			# if translation is executed at the client side 
			my $routingNATFilterList = $addressTranslationParser->routingNAT('automatic');
			$out->print_string($routingNATFilterList);	
			# add the phase discriminator
			$in->{NAT}->{settings}->{processingPhase} = 'routing';
			$in->{NAT}->{settings}->{processingPhasePosition} = 'before';
			my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
			$out->print_string($phaseDiscriminatorPhase);
		$out->close_element('filterListsSection');		
	} else {
		$out->open_element('filterListsSection');
			# else translation is executed at the server side
			my $routingNATFilterList = $addressTranslationParser->routingNAT('automatic');
			$out->print_string($routingNATFilterList);
			# add the phase discriminator
			$in->{NAT}->{settings}->{processingPhase} = 'routing';
			$in->{NAT}->{settings}->{processingPhasePosition} = 'after';
			my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
			$out->print_string($phaseDiscriminatorPhase);
		$out->close_element('filterListsSection');
	}
	
	#check if manual NAT is defined
	$addressTranslationParser->checkManualNAT();
	if (defined ($in->{manualNAT}->{available}) and $in->{manualNAT}->{available} == 1){
		#check where the manual NAT is applied, at the server side or at the client side
		if ( defined($in->{NAT}->{settings}->{nat_dst_client_side_manual}) and $in->{NAT}->{settings}->{nat_dst_client_side_manual} eq 'True'){
			$out->open_element('filterListsSection');
				# if translation is executed at the client side
				my $routingNATFilterList = $addressTranslationParser->routingNAT('manual');
				$out->print_string($routingNATFilterList);	
				# add the phase discriminator
				$in->{NAT}->{settings}->{processingPhase} = 'routing';
				$in->{NAT}->{settings}->{processingPhasePosition} = 'before';
				my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
				$out->print_string($phaseDiscriminatorPhase);		
			$out->close_element('filterListsSection');	
		} else {
			$out->open_element('filterListsSection');
				# else translation is executed at the server side
				my $routingNATFilterList = $addressTranslationParser->routingNAT('manual');
				$out->print_string($routingNATFilterList);
				# add the phase discriminator
				$in->{NAT}->{settings}->{processingPhase} = 'routing';
				$in->{NAT}->{settings}->{processingPhasePosition} = 'after';
				my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
				$out->print_string($phaseDiscriminatorPhase);	
			$out->close_element('filterListsSection');
		}	
	} # check manual NAT existence
		
	
}

sub parse_CPSMS_interfaces {
	
	my ( $in, $out ) = @_;
	
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# create a new parser 
	my $parserno = XML::LibXML->new();
	
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");

	# get gateway's host name, interfaces
	my $gateway = _get_gateway($XPCNO, $ip_addr);
	
	$out->open_element ('interfaces');
		if (defined $gateway->{intfs})	{
			foreach my $intf (@{$gateway->{intfs}}) {
				$out->open_element('interface');
					$out->print_element ('adminStatus','up');
					$out->print_element('ifIndex',$intf->{ifIndex});
						$out->open_element ('interfaceIp');
							my $ref = create_network_element($intf->{ipaddr}, $intf->{netmask});
							$out->open_element('ipConfiguration');
								$out->print_element('ipAddress', $ref->{address});
								$out->print_element('mask', $ref->{mask});
							$out->close_element('ipConfiguration');
						$out->close_element('interfaceIp');
						my $physical;
						if ($intf->{officialname} =~ /eth/){
							$out->print_element('interfaceType','ethernet');
							$physical = 'true';
						} elsif ($intf->{officialname} =~ /lo/){
							$out->print_element('interfaceType','softwareLoopback');
							$physical = 'false';
						} else {
							$out->print_element('interfaceType','virtual');
							$physical = 'false';
						}
						$out->print_element('name',$intf->{officialname});
						$out->print_element('physical', $physical);
				$out->close_element('interface');
			}
		}
		else  { # gateway interfaces not defined
			# We have to create a single interface here since we lack the detail we need to place this.
			# We have the ip addr to assign to it, but everything else must be cobbled together.
			# I hope to be able to remove this code someday when we understand better what we can back
			# up, and what discovering nodes like this really means.
			$LOGGER->debug("parse_CPSMS_interfaces; synthesizing single interface for " .
				$ip_addr . " because interface detail not present in network_objects");

			$out->open_element('interface');
			$out->print_element ('adminStatus','up');
			$out->open_element ('interfaceIp');
			my $ref = create_network_element($ip_addr, "255.255.255.0"); # I know, I know....
			$out->open_element('ipConfiguration');
			$out->print_element('ipAddress', $ref->{address});
			$out->print_element('mask', $ref->{mask});
			$out->close_element('ipConfiguration');
			$out->close_element('interfaceIp');
			$out->print_element('interfaceType','ethernet');
			$out->print_element('name',FAUX_SINGLE_INTERFACE_NAME);						
			$out->print_element('physical', "true");
			$out->close_element('interface');
		}
	$out->close_element('interfaces');
	
}

sub parse_static_routes 
{
	
	my ( $in, $out ) = @_;
	
	my $staticRoutes;
	# get and store CP statuses.xml
	$in->{xml} = "statuses";  
	$in->{statuses} = _cp_to_doc($in);
	
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	
	# get device's ip address 
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# create a new parser 
	my $parserno = XML::LibXML->new();
	
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");

	# get gateway's host name, interfaces
	my $gateway = _get_gateway($XPCNO, $ip_addr);
	_getClusteringInfo($XPCNO, $gateway);
	
	# create a new parser 
	my $parserst = XML::LibXML->new();
	
	# create a new XPath context and register namespace for statuses
	my $XPCST = _xpcregistration($parserst,$in->{statuses},"statuses");
	
	# add static routes info to gateway
	_getStaticRoutes ($XPCST, $gateway);	
	
	if (defined $gateway->{osRouting})
	{
		foreach my $osRouting (@{$gateway->{osRouting}}){
			if ($osRouting->{osRoutingDest} ne '127.0.0.0'
				&& $osRouting->{osRoutingDest} ne '127.0.0.1' 
				&& $osRouting->{osRoutingGateWay} ne '0.0.0.0') {
					my $route = {
						destinationAddress => $osRouting->{osRoutingDest},
						destinationMask    => mask_to_bits($osRouting->{osRoutingMask}),
					};
					$route->{gatewayAddress} = $osRouting->{osRoutingGateWay};
					$route->{interface}      = $osRouting->{osRoutingIntrf};
					if ( ( $route->{destinationAddress} eq '0.0.0.0' ) && ( $route->{destinationMask} eq '0' ) ){
						$route->{defaultGateway} = 'true';
					}
					else{
						$route->{defaultGateway} = 'false';
					}
					push( @{ $staticRoutes->{staticRoute} }, $route );	
			} # end for if osRouting
		} # end foreach osRouting
	}
	else {
		$LOGGER->debug("parse_static_routes; no static routes defined");
	}
	
	$out->print_element( 'staticRoutes', $staticRoutes ) if ( defined $staticRoutes );
		
}

sub parse_snmp {
	
	my ($in, $out) = @_;
	
	$in->{xml} = "network_objects";
	$in->{snmp} = _cp_to_doc($in);
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# create a new parser 
	my $parser = XML::LibXML->new();
	
	# create a new XPath context and register namespace
	my $XPC = _xpcregistration($parser,$in->{snmp},"snmp");
	
	# find SNMP settings for the given ip_addr 
	foreach my $libxmlobject ($XPC->findnodes('//object')){
		my $ipaddr = ($libxmlobject->findvalue('./ipaddr'));
		if ($ipaddr eq $ip_addr ) {
			# find object node from SNMP
			foreach my $libxmlSNMP ($libxmlobject->findnodes('./SNMP')) {
				foreach my $libxmlObj ($libxmlSNMP->findnodes('./object')){
					#find values for a non-empty sysName values
					my $sysName = $libxmlObj->findvalue('./sysName');
					my %community;
					if(defined ($sysName) && ($sysName ne "")){
						$out->open_element('snmp');
							#search for read_community
							my $read_community = $libxmlObj->findvalue('./read_community');
							if (defined ($read_community) && ($read_community ne "")){
								my $accessType = "RO";
								my $communityString = $read_community;
								$community{$communityString}=$accessType;
							}
							
							#search for readwrite_community
							my $write_community = $libxmlObj->findvalue('./write_community');
							if (defined ($write_community) && ($write_community ne "")){
								my $accessType = "RW";
								my $communityString = $write_community;
								$community{$communityString}=$accessType;
							}
							
							#print community element 
							foreach my $key (keys %community){
								$out->open_element('community');
									$out->print_element('accessType',$community{$key});
									$out->print_element('communityString',$key);
								$out->close_element('community');
							}
							
							#sysContact
							my $sysContact = $libxmlObj->findvalue ('./sysContact');
							if(defined ($sysContact) && ($sysContact ne "")){
								$out->print_element('sysContact', $sysContact);
							}	
							
							#sysDescription
							my $sysDescr = $libxmlObj->findvalue ('./sysDescr');
							if(defined ($sysDescr) && ($sysDescr ne "")){
								$out->print_element('sysDescr', $sysDescr);
							}
							
							#sysLocation
							my $sysLocation = $libxmlObj->findvalue ('./sysLocation');
							if(defined ($sysLocation) && ($sysLocation ne "")){
								$out->print_element('sysLocation', $sysLocation);
							}
							
							$out->print_element('sysName', $sysName);
						$out->close_element('snmp');		
					} #end if sysName
				} #end foreach libxmlObj	
			} #end foreach libxmlSNMP
								
		} #end if ipaddr
	} #end foreach libxmlobject	
			
}

sub _xpcregistration {
	
	my ($parser, $in, $ns) = @_;
	
	# create a new DOM
	my $dom = $parser->parse_string($in);
	# set document's root
	my $doc = $dom->documentElement();
	# create a new XPath context and register namespace
	my $XPC     = XML::LibXML::XPathContext->new($doc);
	$XPC->registerNs( $ns => 'http://www.ziptie.org/model/common/1.0' );
	
	return $XPC;
	
}

sub _printnodes{
	
	my ($doc, $in, $out) = @_;
	
	my $ns = undef;
	my $item = undef;
	
	if (($in !~ /^core/) && ($in =~ /\:/)){
		($ns, $item) =split (':',$in);
	} else {
		$item = $in;
	}
	
	# find core:x nodes
	foreach my $libxmlvar ($doc->findnodes('//'.$in)) {
		if (defined $libxmlvar and $libxmlvar ne ''){
			my $xmlinvar = $libxmlvar->toString;
			my $ref = XMLin($xmlinvar, ForceArray =>1);
			$out->print_element($item, $ref);	
		} else {
			$out->print_element($item, 'unknown');
		}		
	}
}

sub _printelement{
	
	my ($doc, $in, $out) = @_;
	my $ns = undef;
	my $item = undef;
	
	if (($in !~ /^core/) && ($in =~ /\:/)){
		($ns, $item) =split (':',$in);
	} else {
		$item = $in;
	}
	
	# find core:x element's value
	foreach my $libxmlvar ($doc->findvalue('//'.$in)){
		if (defined $libxmlvar and $libxmlvar ne ''){
			my $ref = $libxmlvar;
			$out->print_element($item, $ref);	
		} else {
			$out->print_element($item, 'unknown');
		}
		
	} 
}

sub _cp_to_sed_xml
{
	my ($in,$xp_params) = @_;
	
	# create a new CPSMSConverter session
	my $cptxml = ZipTie::Adapters::CheckPoint::CPSMS::CPSMSConverter->new(connectionPath => $in->{cpsms_ctx}->{connectionPath},
																			cpt_dir => $in->{cpsms_ctx}->{cpt_dir},
																			xml_dir => $in->{cpsms_ctx}->{xml_dir},
																			xsl_dir => $in->{cpsms_ctx}->{xsl_dir},
																			xml => $in->{xml},
																		    xsl => $in->{xsl},
																		    xp_params => $xp_params);
	my $result = $cptxml->cp2sedxml();
	return $result;																		    															   		
}

sub _cp_to_sed_xmlstring
{
	my ($in,$xp_params) = @_;
	
	my $xml = exists $in->{xml} ? $in->{xml} : "";
	# create a new CPSMSConverter session
	my $cptxml = ZipTie::Adapters::CheckPoint::CPSMS::CPSMSConverter->new(connectionPath => $in->{cpsms_ctx}->{connectionPath},
																			cpt_dir => $in->{cpsms_ctx}->{cpt_dir},
																			xml_dir => $in->{cpsms_ctx}->{xml_dir},
																			xsl_dir => $in->{cpsms_ctx}->{xsl_dir},
																			xml => $xml,
																			xmlstring => $in->{xmlstring},
																		    xsl => $in->{xsl},
																		    xp_params => $xp_params);
	my $result = $cptxml->cp2sedxmlstring();
	return $result;																		    															   		
}

sub _cp_to_doc
{
	my ($in,$xp_params) = @_;
	
	# create a new CPSMSConverter session
	my $cptxml = ZipTie::Adapters::CheckPoint::CPSMS::CPSMSConverter->new(connectionPath => $in->{cpsms_ctx}->{connectionPath},
																			cpt_dir => $in->{cpsms_ctx}->{cpt_dir},
																			xml_dir => $in->{cpsms_ctx}->{xml_dir},
																			xsl_dir => $in->{cpsms_ctx}->{xsl_dir},
																			xml => $in->{xml},
																		    xsl => $in->{xsl},
																		    xp_params => $xp_params);
	my $result = $cptxml->cpxml2doc();
	return $result;																		    															   		
}

# returns substring that is found after a passed string
sub _substring_after
{
	my ($in,$after) = @_ ;
	my $result;
	if ($in =~ /$after(.*)/){
		$result = $1;
	}
	return $result;
}

# set objectGroupReference
sub _set_objectGroupReference{
	
	my ($in) = @_;
	my @obj;
	push (@obj ,{objectGroupReference => $in});
	return @obj;
	
}

# Gets data on a gateway specified by an ip address of one of its interfaces
sub _get_gateway {
	my ($XPCNO, $gateway_ip) = @_;
	my $gateway;

	GATEWAY_SEARCH:	# Search through each network object in the raw config for the gateway ip
	for my $libxmlobject ( $XPCNO->findnodes('/network_objects/object') ) {

		# Skip unless this is a host object with a firewall installed / a cluster member
		my $firewall_installed = $libxmlobject->findvalue('./firewall');
		my $type = $libxmlobject->findvalue('./type');
		next unless $firewall_installed eq 'installed' or $type eq 'cluster_member';

		my $hostname = $libxmlobject->findvalue('./object_name');
		my $ip = $libxmlobject->findvalue('./ipaddr');
		my $ip_match = $gateway_ip =~ /^$ip$/;

		my @interfaces;
		for my $libxmlinterface ( $libxmlobject->findnodes('./interfaces/object') ) {
			my $interface_ip = $libxmlinterface->findvalue('./ipaddr');
			my $interface = {
				ipaddr       => $interface_ip,
				ifIndex      => $libxmlinterface->findvalue('./ifindex'),
				netmask      => $libxmlinterface->findvalue('./netmask'),
				officialname => $libxmlinterface->findvalue('./officialname')
			};

			# Check if this interface is external or internal
			for my $libxmltypeintf ( $libxmlinterface->findnodes('./security/object/netaccess/object') ) {
				my $typeofIntf = $libxmltypeintf->findvalue('./leads_to_internet');
				$interface->{typeofIntf} = $typeofIntf eq 'True' ? 'external' : 'internal';
			}

			push @interfaces, $interface;

			# Management IP may be on a nested interface object
			$ip_match = 1 if $gateway_ip =~ /^$interface_ip$/;
		}

		# Next object unless the IP matches the parameter
		next unless $ip_match;

		$gateway = {
			hostname                => $hostname,
			my_ip                   => $ip,
			my_ip_and_local_cluster => { my_ip => $ip },
			intfs                   => \@interfaces
		};

		last;
	}

	return $gateway;
}

# get CP osRouting information from statuses.xml and collect them under gateway->{osRouting}
sub _getStaticRoutes {
	
	my ($XPCST, $gateway) = @_;
	
	my $hostname = $gateway->{hostname};
	
	# need to use cluster member hostname if gateway cluster to find routes
	if ($gateway->{isGatewayCluster}){
		if ($gateway->{cluster_members} > 0){
			$hostname = $gateway->{cluster_members}[0];
		}
	}
		
	# find and parse application_obj and osRoutingInfo for a given hostname 
	foreach my $libxmlobject ($XPCST->findnodes('/statuses/object')){
		my $appObj = $libxmlobject->findvalue('./application_obj');
		$appObj= _substring_after($appObj,"referenced_object: ");
		# find the matching hostname cpshared_application
		if ($appObj =~ /cpshared_application_$hostname/){
			foreach my $libxmlOsRoute ($libxmlobject->findnodes('./osStatus/object/osRoutingInfo/object/osRoutingEntry/object')) {
				# create an empty bucket for routing
				my $osRouting ={};	
				$osRouting->{osRoutingDest} = $libxmlOsRoute->findvalue('./osRoutingDest');
				$osRouting->{osRoutingGateWay} = $libxmlOsRoute->findvalue('./osRoutingGateWay');
				$osRouting->{osRoutingIntrf} = $libxmlOsRoute->findvalue('./osRoutingIntrf');
				$osRouting->{osRoutingMask} = $libxmlOsRoute->findvalue('./osRoutingMask');
				# push routing information under osRouting bucket 
				push (@{$gateway->{osRouting}}, $osRouting);
			} # end foreach libxmlOsRoute  		
		} # end of if find matching hostname 
	} #end foreach libxmlobject	
}

# get CP MemoryStat-uses and collect them under gateway->{MemoryStat}
sub _getMemoryStat {
	my ($XPCST,$gateway) = @_;
	
	my $hostname = $gateway->{hostname};
	# find and parse application_obj and osRoutingInfo for a given hostname 
	foreach my $libxmlobject ($XPCST->findnodes('/statuses/object')){
		my $appObj = $libxmlobject->findvalue('./application_obj');
		$appObj= _substring_after($appObj,"referenced_object: ");
		# find the matching hostname cpshared_application
		if ($appObj =~ /cpshared_application_$hostname/){
			foreach my $libxmlMemoryStat ($libxmlobject->findnodes('./osStatus/object/osPerfStat/object/MemoryStat/object')) {
				# create an empty bucket for MemoryStat
				my $MemoryStat ={};	
				$MemoryStat->{totalVirtMem} = $libxmlMemoryStat->findvalue('./total_virtual_mem');
				push (@{$gateway->{MemoryStat}}, $MemoryStat);
			} # end foreach	
		} # end of if find matching hostname
	} # end of foreach libxmlobject	
}

sub _getModel {
	my ($XPCNO,$XPCST, $gateway) = @_;
	
	my $version;
	my $hostname = $gateway->{hostname};
	# need to use gateway cluster hostname if cluster member
	if ($gateway->{isClusterMember}){
		if ($gateway->{gateway_clusters} > 0){
			$hostname = $gateway->{gateway_clusters}[0];
		}
	}
	
	foreach my $libxmlobject ($XPCNO->findnodes('/network_objects/object')){
		my $network_object_name = $libxmlobject->findvalue('./object_name');
		$version = $libxmlobject->findvalue('./svn_version_name');
		if ($network_object_name eq $hostname){
			$version = $libxmlobject->findvalue('./svn_version_name');
			if ($version ne ""){
				return $version;
			}
		}
	}
	# if version still not found use statuses
	# find and parse application_obj and osRoutingInfo for a given hostname 
	foreach my $libxmlobject ($XPCST->findnodes('/statuses/object')){
		my $appObj = $libxmlobject->findvalue('./application_obj');
		$appObj= _substring_after($appObj,"referenced_object: ");
		# find the matching hostname cpshared_application
		if ($appObj =~ /cpshared_application_$hostname/){
			$version = $libxmlobject->findvalue('./osStatus/object/CPsharedInfo/object/CPsharedVersion');
			last;
		}	
	}
	return $version; 	
}

# List of exceptions names for positional properties
my %AlternateServicePositionalPropertyMappings = (
		# R75.40 Gaia
		"r_access_enable_p" => "raccessenable", 
		#"allow_gw_local_services_p" => "allow_gw_local_services"
);

# Function checks properties enablement.
# It returns a hash variable containing the properties relevant for implied_rules
# A valid property has to be enabled (i.e property = True) and 
# has an associated positional property (i.e property_p = '<first> <before last> <last>')  
sub _getPropertiesEnablement {

	my ($XPCPR, $property) = @_;
	my $properties = {};
	
		foreach my $libxmlobject($XPCPR->findnodes('/properties/object')){
			my $objname = $libxmlobject->findvalue('./object_name');
			if ($objname =~ /firewall_properties/){
				if (defined $property){
				# check if $property is available. The found property is a good candidate to build unordered rules 
					my $enProperty = undef;  	
					$enProperty = $libxmlobject->findvalue("./$property");	
					if (defined ($enProperty) and ($enProperty eq 'True')){
						$properties->{$property}->{enabled} = 1;
					}			
				}else {
					# find all properties that have a positional property
					foreach my $posProperty ($libxmlobject->findnodes("./*[contains(local-name(), '_p')]")) {
						my $propertyName = undef;
						my $valProperty = undef;
						# node's name
						my $nodeName = $posProperty->nodeName;
						if ($nodeName =~ /(^\w+)_enable_p$/) {
							# check for special property names
							my $property = $AlternateServicePositionalPropertyMappings{$nodeName};
							if (defined ($property) and ($property ne "") ){
								$propertyName = $property;
							}else {
								$propertyName = $1;
							}
						} elsif ($nodeName =~ /(^\w+)enable_p$/){
							$propertyName = $1;
						} elsif ($nodeName =~ /(^\w+)_p$/){
							$propertyName = $1; 
						# could not find a matching property corresponding to the positional property		
						} else {
							$propertyName = $AlternateServicePositionalPropertyMappings{$nodeName};
						} 
						
						if (defined ($propertyName)){
							# alternative possible property name
							my $altPropertyName = $propertyName."enable";
							# search for properties associated with the positional properties 
							my $enProperty = undef;
							$valProperty = $libxmlobject->findvalue("./$altPropertyName");
							if ((defined $valProperty) and ($valProperty eq 'True')){
								$enProperty = $altPropertyName; 
							}
							$valProperty = $libxmlobject->findvalue("./$propertyName");
							if ((defined $valProperty) and ($valProperty eq 'True')) {
								$enProperty = $propertyName;	
							}		
							# check if property is enabled
							if (defined ($enProperty)){
								$properties->{$enProperty}->{enabled} = 1;
								# positional's node value
								$properties->{$enProperty}->{position} = $posProperty->findvalue(".");
							}
						} #end of if defined propertyName		
						
					} #end foreach posProperty
				} # end else	
			} # end 
		} # end foreach $XPCPR
		
	return $properties;	
}

# process properties and builds $impliedFeSets that is going to be feed 
# to generate_filter_list 	
 sub _processProperties{
 	my ($XPCPR, $prop,$prop_with_pos) = @_;
 	
 	my $impliedFeSets = {};
 	# parse property names from impliedFeSets file and compare them
 	# with the properties that are enabled and have a defined positional property
 	for my $propName (keys %{$prop}) {
 		my $struct =  ${$prop}{$propName};
       	    	
    	for my $posProp (keys %{$prop_with_pos}) {
    		if  ($propName eq $posProp ){ 
	    		# push-in the ordered rules
	    		my $position = ${$prop_with_pos}{$posProp}{position};
	    		foreach my $valofStruct (@{$struct}){
	    			push (@{$impliedFeSets->{$position}}, $valofStruct);	
	    		}
    		} 	 	
    	} #end for $posProp
    	
    	my $property = _getPropertiesEnablement($XPCPR,$propName);
    	if (($property->{$propName}->{enabled}) and ( not defined(${$prop_with_pos}{$propName}{position})) ){
    		foreach my $valofStruct (@{$struct}){
    			push (@{$impliedFeSets->{'unordered'}}, $valofStruct);
    		}
    	} #end if $property->{$propName}->{enabled}
        	
    }#end for $propName
        
 	return $impliedFeSets;	
 	
 }

sub _prune_global_rules {
	my $fPruneList = shift;
	my $serviceGroups = shift;
	
	if (not defined $fPruneList) {
		return;
	}
	
	for (my $pruneIdx = scalar(@{$fPruneList} - 1); $pruneIdx >= 0; $pruneIdx--) {
		my $fPruneEnt = $fPruneList->[$pruneIdx];
		my $anyAddrEnt = undef;
		
		# (re) assign process order
		$fPruneEnt->{processOrder} = $pruneIdx + 1;
		
		# Check for a source, dest of any, if it doesn't match, we can move to the next 
		# filter entry	
		foreach my $xAddrList ($fPruneEnt->{sourceIpAddr}, $fPruneEnt->{destinationIpAddr}) {
			for (@{$xAddrList}) {  # check for not qualified by object group ref and address is any addr
				if ( (not defined $_->{objectGroupReference}) and
					($_->{network}->{address} eq get_any_address_address()))	{
					$anyAddrEnt = $_;
					last;
				}
			}
			
			if (defined $anyAddrEnt) {
				last;
			}									
		}
		# did we match in src or dest?
		if (not defined $anyAddrEnt) {
			# yes.  There's no matching wilcard-any addr, so let's leave this alone and move on to the next entry
			next; 
		}
		
		# if protocol for src or dest isn't specified, then delete
		if ( (not defined $fPruneEnt->{protocol}) and
			(not defined $fPruneEnt->{destinationService}) and
			(not defined $fPruneEnt->{sourceService})) {
				# delete at this list entry.  Be careful to delete from the array, not the
				# entry itself, which would just assign an undef to this slot.
				splice (@{$fPruneList}, $pruneIdx, 1);				
		}

		foreach my $destService ($fPruneEnt->{destinationService}){
			foreach my $destServiceEntry (@{$destService}){
				if (defined $destServiceEntry->{objectGroupReference}){
					my $destServiceRef = $destServiceEntry->{objectGroupReference};
					unless (defined $serviceGroups->{$destServiceRef}){
						splice (@{$fPruneList}, $pruneIdx, 1);
						$LOGGER->debug("_prune_global_rules; Implied rule with missing service being DROPPED"); 
						last;
					}
				}
			}
		}
	}
}
 
# Function parses services.xml and search for src_port definitions
# if it is available it will push-in a new serviceGroup__src structure   
sub _createSrcPortServiceGroup {
 	my ($id, $XPCSRV) = @_;
 	my %source_services = ();

 	foreach my $service_object ($XPCSRV->findnodes('/services/object')){
		my $name = $service_object->findvalue('./object_name');
		if ($id eq $name){
			my $srcPort = $service_object->findvalue('./src_port');
			my $protocol = lc $service_object->findvalue('./type');
			if( $srcPort ){
				if ( $srcPort =~ /^\d+$/ ){
					push @{ $source_services{$id . '__src'}->{portExpression} }, {
						port     => $srcPort,
						operator => 'eq',
						protocol => $protocol
					};
				} elsif ( $srcPort =~ /^(\d+)\s?-\s?(\d+)$/ ) {
					push @{ $source_services{$id . '__src'}->{portRange} }, {
						portStart => $1,
						portEnd   => $2,
						protocol  => $protocol
					};
				}
			}
		}
 	}
	return \%source_services;
 }
 
 sub _checkServiceGroup {
 	my ($in, $service, $filterEntry) = @_;
	my %protocols;

	for my $service_direction ( $service, $service.'__src' ) {
		if( defined $in->{parsed_services}->{$service_direction} ) {
			my $model_direction = $service_direction eq $service ? 'destinationService' : 'sourceService';
			push @{$filterEntry->{$model_direction}}, $service_direction;

			my $service_protocols = _get_service_protocols( $in, $service_direction );
			%protocols = ( %protocols, %{ $service_protocols } );
		}
	}
	my @protocol_names = keys %protocols;
	if( @protocol_names ) {
		#$LOGGER->debug("Protocol(s) found for '$service': @protocol_names");
		$filterEntry->{protocol} = {} if not defined $filterEntry->{protocol};
		%{ $filterEntry->{protocol} } = ( %{ $filterEntry->{protocol} }, %protocols );
	} else {
		#$LOGGER->debug("No protocols found for service '$service'. Ommiting rule protocol designation.");
	} 

	return;
}

# Recursively searches a service object to find all the protocols the services, and thus rule, apply to
sub _get_service_protocols {
	my ($in, $service) = @_;
	my %protocols = ();

	my $service_model = $in->{parsed_services}->{$service};
	for my $key ( %{ $service_model } ) {
		if( $key eq 'objectGroupReference' ) {
			for my $ref ( @{ $service_model->{objectGroupReference} } ) {
				my $ref_protocols = _get_service_protocols($in, $ref);
				%protocols = ( %protocols, %{ $ref_protocols } );
			}
		} elsif( $key ne 'id' ) {
			for my $expression_model ( @{ $service_model->{$key} } ) {
				if( $expression_model->{protocol} ) {
					$protocols{$expression_model->{protocol}} = 1;
				}
			}
		}
	}

	return \%protocols;
 }
 
# parse network_objects to check if NAT is defined.
# looking after firewalling_obj elements and store the NAT elements of the network objects
sub _checkNAT {
        my ($in, $XPCNO, $XPCFW, $gateway) = @_;
        $in->{isNAT} = 0;

        # check if automatic NAT is defined 
        foreach my $libxmlobject ($XPCNO->findnodes('/network_objects/object')){
                my $object_name = $libxmlobject->findvalue('./object_name');
                        foreach my $libxmlNAT($libxmlobject->findnodes('./NAT')) {
                                foreach my $libxmlNATObj ($libxmlNAT->findnodes('./object')){
                                        my $nat;

                                        # store NAT settings for future references 
                                        # NAT rules are installed on the Gateway specified in the_firewalling_obj
                                        $nat->{the_firewalling_obj} = $libxmlNATObj->findvalue('./the_firewalling_obj');

                                        # store NAT translation method (i.e static or hide (dynamic) )
                                        $nat->{netobj_adtr_method} = $libxmlNATObj->findvalue('./netobj_adtr_method');
                                        
                                        # valid address name
                                        $nat->{valid_addr_name} = $libxmlNATObj->findvalue('./valid_addr_name');

                                        # valid NAT IP address 
                                        $nat->{valid_ipaddr} = $libxmlNATObj->findvalue('./valid_ipaddr');
                                        
                                        # found a network_objects that has a reference to the analyzed gateway
                                        # NAT rules can be installed on a specific gateway or on all available managed gateways
                                        if( not $in->{isNAT} ) {
											if( defined $nat->{the_firewalling_obj} ) {
                                            	if( ( $nat->{the_firewalling_obj} =~ /All|Gateways/ ) or ( $gateway->{hostname} and $nat->{the_firewalling_obj} =~ /$gateway->{hostname}/ ) ) {
													$in->{NAT}->{installed_on_gateway} = $gateway;
													$in->{isNAT} = 1;
												}
                                            }
                                        }
                                        $in->{NAT}->{$object_name} = $nat;
                                }
                        }
        }

        # check if manual NAT is defined
        for my $libxmlobject ( $XPCFW->findnodes('/fw_policies/object/rule_adtr/rule_adtr/object') ) {
                for my $libxmlinstall ( $libxmlobject->findnodes('./install') ) {
                        my $install = $libxmlinstall->findvalue('./install');
                        if ( not $in->{isNAT} ) {
				if( defined $install ) {
					if( ( $install = /Any|Gateways/ ) or ( $gateway->{hostname} and $install =~ /$gateway->{hostname}/ ) ) {
						$in->{NAT}->{installed_on_gateway} = $gateway;
						$in->{isNAT} = 1;
					}
                                }
                        }
                }
        }

        return;
}

sub parse_network_groups {
	my ($out, $XPC, $ip_addr, $in, $XPCNO) = @_;

	# Due to the xslt legacy there are two places network objects are parsed from
	# The xslt'd config in $XPC and the raw config in $XPCNO
	# $XPC network groups are needed to check references in $XPCNO
	my $xslt_network_groups = _parse_network_groups_xslt($XPC, $XPCNO, $ip_addr);
	my $raw_network_groups = _parse_network_groups_raw($XPCNO, $xslt_network_groups);

	my %network_groups = %{ $raw_network_groups };
	_print_network_groups($out, \%network_groups);

	# Store for checking references during rule parsing
	$in->{networkGroups} = \%network_groups;

	return;
}

sub _parse_network_groups_xslt {
	my ($XPC, $XPCNO, $ip_addr) = @_;
	my %network_groups = ();

	# find and print networkGroup node
	foreach my $libxmlNG ( $XPC->findnodes('//ngns:networkGroup') ) {
		my $id = $XPC->findvalue('.//ngns:id', $libxmlNG);
		my $network_group = { id => $id };

		# find host elements
		my $host = $XPC->findvalue('.//ngns:host', $libxmlNG);
		if( $host ) {
			push @{ $network_group->{host} }, $host;

			# Make the builtin network objects for this gateway
			if ($host eq $ip_addr){
				my $gateway = _get_gateway($XPCNO, $host);

				if( $gateway) {
					$network_groups{'my_ip'} = { id => 'my_ip', host => $gateway->{my_ip} };
					$network_groups{'my_ip_and_local_cluster'} = { id => 'my_ip_and_local_cluster', host => $gateway->{my_ip} };
				}
			}
		}

		# find subnet elements
		foreach my $libxmlsub ($XPC->findnodes('.//ngns:subnet',$libxmlNG)) {
			my $xmlinvar = $libxmlsub->toString;
			my $ref = XMLin($xmlinvar, ForceArray => 1);
			push @{ $network_group->{subnet} }, $ref;	
		}
		
		# find objectGroupReferences
		foreach my $libxmlsub ($XPC->findnodes('.//ngns:objectGroupReference',$libxmlNG)) {
			my $xmlinvar = $libxmlsub->toString;
			if( $xmlinvar eq '<objectGroupReference/>' ) {
				next;
			}
			my $ref = XMLin($xmlinvar, ForceArray => 1);
			push @{ $network_group->{objectGroupReference} }, $ref;
		}
		
		# find IP ranges (or IP pools)
		foreach my $libxmlsub ($XPC->findnodes('.//ngns:range',$libxmlNG)) {
			my $xmlinvar = $libxmlsub->toString;
			my $ref = XMLin($xmlinvar, ForceArray => 1);
			push @{ $network_group->{range} }, $ref;
		}

		$network_groups{$id} = $network_group;
	}

	return \%network_groups;
}

sub _parse_network_groups_raw {
	my ($XPCNO, $network_groups_ref) = @_;
	my %network_groups = %{ $network_groups_ref };

	for my $node ( $XPCNO->findnodes('/network_objects/object') ) {
		my $id = $node->findvalue('./object_name');
		my $network_group = { id => $id };
		my $type = $node->findvalue('./type');

		if( $type eq 'group_with_exclusion' ) {
			my $base = $node->findvalue('./base');
			my $base_set;

 			$base =~ s/referenced_object:\s//;
			if( $base =~ /any/i ) {
				$base_set = Net::CIDR::Set->new('0.0.0.0/0');
			} else {
				$base_set = _get_network_set($base, \%network_groups);
			}

			if( $base_set ) {
				my $diff_set = $base_set;

				# Get addresses in base_set but not in exception_set
				# unless the exception set is empty, then use the base
				my $exception = $node->findvalue('./exception');
				$exception =~ s/referenced_object:\s//;
				my $exception_set = _get_network_set($exception, \%network_groups);
				if( $exception_set ) {
					$diff_set = $base_set->diff($exception_set);
				}

				my @diffs = $diff_set->as_range_array();
				for my $diff( @diffs ) {
					if( $diff =~ /(.+)-(.+)/ ) {
						push @{ $network_group->{range} }, { startAddress => [$1], endAddress => [$2] }; # must store range values as arrays to align with XSLT parsing
					} else {
						my ($address, $mask) = split /\//, $diff;
						push @{ $network_group->{subnet} }, { address => [$address], mask => [$mask] // [32] }; # same for subnets
					}
				}
				$network_groups{$id} = $network_group;
			} else {
				$LOGGER->debug("WARNING: Dropping network group. Unknown 'group with exclusion' base for $id: '$base'");
			}
		} elsif( $type eq 'group' ) {
			for my $child_node ( $node->findnodes( './empty_field_name/empty_field_name' ) ) {
				my $reference = $child_node->findvalue('.');
				$reference =~ s/referenced_object:\s//;
				if( $reference ne 'no_elements_available' ) {
					if( $network_groups{$reference} ) { 
						push @{ $network_group->{objectGroupReference} }, $reference;
					} else {
						$LOGGER->debug("WARNING: Dropping network group. Unknown 'group' reference for $id: '$reference'");
					}
				}
			}
			$network_groups{$id} = $network_group;
		} elsif( $type eq 'voip_sip' ) {
			# if already exists delete it
			if ($network_groups{$id}){
				delete $network_groups{$id};
				$LOGGER->debug("WARNING: Dropping network group $id. VOIP SIP types not supported");
			}
		}
	}

	return \%network_groups;
}

sub _print_network_groups {
	my ($out, $network_groups_ref) = @_;
	my %network_groups = %{ $network_groups_ref };

	for my $id ( keys %network_groups ) {
		my $network_group = $network_groups{$id};
		$out->open_element('networkGroup');
		$out->print_element('id', $id);
		for my $element ( qw(host objectGroupReference range subnet) ) {
			$out->print_element($element, $network_group->{$element}) if $network_group->{$element};
		}
		$out->close_element('networkGroup');
	}
	return;
}

sub _get_network_set {
	my ($name, $network_groups) = @_;
	my $set;

	if( exists $network_groups->{$name} ) {
		my %network_group = %{ $network_groups->{$name} };
		$set = Net::CIDR::Set->new();
		for my $key ( keys %network_group ) {
			my $value = $network_group{$key};
			if( $key eq 'objectGroupReference' ) {
				for my $ref ( @{ $value } ) {
					my $nested_set = _get_network_set($ref, $network_groups);
					if( $nested_set ) {
						$set->merge($nested_set);
					}
				}
			} elsif( $key eq 'host' ) {
				for my $host ( @{ $value } ) {
					$set->add($host);
				}
			} elsif( $key eq 'subnet' ) {
				for my $subnet ( @{ $value } ) {
					my $ip = $subnet->{address}[0];
					my $mask = $subnet->{mask}[0];
					$set->add("$ip/$mask");
				}
			} elsif( $key eq 'range' ) {
				for my $range ( @{ $value } ) {
					my $start = $range->{startAddress}[0];
					my $end = $range->{endAddress}[0];
					$set->add("$start-$end");
				}
			}
		}
	}

	return $set;
}

sub parse_protocol_groups {
	my $out = shift;
	my $XPC = shift;

	foreach my $libxmlPG (
		$XPC->findnodes(
			'//psgns:protocolGroup')) {
		my $id = $XPC->findvalue('.//psgns:id', $libxmlPG);
	
		foreach my $protocol ($XPC->findnodes('.//psgns:protocol',$libxmlPG)){
			if (defined ($protocol) && ($protocol) ne ""){
				my $xmlinvar = $protocol->toString;
				next if ($xmlinvar =~ /<protocol\/>/);
				my $ref = XMLin($xmlinvar, ForceArray =>1);
				$out->open_element('protocolGroup');
					$out->print_element('id', $id);
					$out->print_element('protocol', $ref);
				$out->close_element('protocolGroup');
			}	
		} # end foreach protocol
	
		my $objPGR = 0;
		foreach my $objectGroupReference($XPC->findnodes('.//psgns:objectGroupReference',$libxmlPG)){
			if (defined ($objectGroupReference) && ($objectGroupReference ne "")){
				my $xmlinvar = $objectGroupReference->toString;
				my $ref = XMLin($xmlinvar, ForceArray =>1);
				if ($objPGR == 0){
					$out->open_element('protocolGroup');
					$out->print_element('id', $id);
				}	
				$out->print_element('objectGroupReference', $ref);
				$objPGR = 1;	
			} #end if		
		} # end foreach protocolGroup objectGroupReference
		if ($objPGR == 1){
			$out->close_element('protocolGroup');
		}				
	} # end foreach protocolGroup
}

sub parse_service_groups {
	my ($out, $services_xml, $in) = @_;
	my %services;
	my %check; # used to check for missing refs

	# Parse each service into %services
	for my $service_object ($services_xml->findnodes('/services/object')){
		my $name = $service_object->findvalue('./object_name');

		# Find the transport protocol (tcp, udp, icmp) or 'group'
		my $type = lc $service_object->findvalue('./type');

		if( $type eq 'group' ) {
			my @groups = $service_object->findnodes('./empty_field_name/empty_field_name');
			for my $group ( @groups ) {
				$group = $group->findvalue('.');
				$group =~ s/referenced_object:\s//;

				if( defined $services{$group} ) {
					push @{ $services{$name}->{objectGroupReference} }, $group;

					# if this service has a src component it is also referenced
					if( defined $services{"$group\__src"} ) {
						push @{ $services{"$name\__src"}->{objectGroupReference} }, "$group\__src";
					}
				} else {
					push @{ $check{$name} }, $group;
				}
			}
		} else {
			my $port = $service_object->findvalue('./port');
			if( not $port ) {
				$port = $service_object->findvalue('./user_port');
			}

			my $protocol = $service_object->findvalue('./protocol');
			my $guessed_port;

			if( $type =~ /(udp|tcp|icmp)/ ) {
				$protocol = $1;
			} elsif( $type =~ /dcerpc/ ) {
				$protocol = 'tcp';
				$guessed_port = 135;
			} elsif( $type =~ /rpc/ ) {
				# in this case the port is actually the rpc application number
				if( $port == 100000 ) {
					$protocol = 'tcp|udp';
					$guessed_port = 111;
				} elsif( $port == 100003 ) {
					$protocol = 'tcp|udp';
					$guessed_port = 2049;
				} elsif( $port == 100005 ) {
					$protocol = 'tcp';
					$guessed_port = 839;
				} else {
					$LOGGER->debug("WARNING: Dropping service '$name'. Unknown RPC application number '$port'");
					next;
				}
			# Use the network protocol to work out the transport protocol
			} elsif( $protocol ) {
				if( $protocol == 55 ) { # mobile-ip uses tcp/434 and udp/434
					$protocol = 'tcp|udp';
					$guessed_port = 434;
				} elsif( $protocol == -1 ) {
					$LOGGER->debug("WARNING: Dropping service '$name'. Application layer services not supported.");
					next;
				} else {
					undef $protocol;
				}
			}

			$services{$name} = _parse_service($port, $protocol, $guessed_port);

			my $src_port = $service_object->findvalue('./src_port');
			if( $src_port ) {
				$services{"$name\__src"} = _parse_service($src_port, $protocol, $guessed_port);
			}
		}
	}

	for my $name ( keys %check ) {
		my $refs = $check{$name};
		for my $ref ( @{ $refs } ) {
			if( not $services{$ref} ) {
				$LOGGER->debug("WARNING: Dropping service '$name'. Unknown refereced object '$ref'");
				delete $services{$name}; last;
			}
		}
	}

	$in->{parsed_services} = \%services;

	my @services_keys = keys %services;
	for my $service ( @services_keys ) {
		my %service_model = ( %{ $services{$service} }, ( id => $service ) );
		$out->print_element('serviceGroup', \%service_model);
	}

	return;
}

sub _parse_service {
	my ($port, $protocol, $guessed_port) = @_;
	my $service = {};

	if( $guessed_port ) {
		my @protocols = split /\|/, $protocol;
		for $protocol ( @protocols ) {
			my $port_expression = {
				operator => 'eq',
				port => $guessed_port,
			};
			if( $protocol ) {
				$port_expression->{protocol} = $protocol;
			}
			push @{$service->{portExpression}}, $port_expression;
		}
	} elsif( $port =~ /^(\d+)-(\d+)$/ ) {
		my $port_range = {
			portStart => $1,
			portEnd => $2
		};
		if( $protocol ) {
			$port_range->{protocol} = $protocol;
		}
		push @{$service->{portRange}}, $port_range;
	} elsif ( $port =~ /^([><])?(\d+)/ ) {
		my $operator = $1 ? $1 eq '>' ? 'gt' : 'lt' : 'eq';
		my $port_expression = {
			operator => $operator,
			port => $2,
		};
		if( $protocol ) {
			$port_expression->{protocol} = $protocol;
		}
		push @{$service->{portExpression}}, $port_expression;
	} elsif( $protocol ) {
		push @{$service->{portRange}}, {
                        portStart => 0,
                        portEnd => 65535,
                        protocol => $protocol
                };
	}

	return $service;
}
