package ZipTie::Adapters::CheckPoint::HTTPS::Parsers;

use strict;
use warnings;
use 5.010_000;

use Data::Dumper;
use Net::CIDR::Set;
use Storable qw(dclone);

use ZipTie::Adapters::CheckPoint::HTTPS::Utils qw(get_server_by_ip get_policy_by_name get_network_group_set get_protocol_for_service);
use ZipTie::Adapters::Utils qw(get_crep mask_to_bits);
use ZipTie::Logger;

use Exporter qw(import);
our @EXPORT_OK = qw(parse_system parse_chassis parse_interfaces parse_object_groups parse_filters parse_static_routing);

my $LOGGER = ZipTie::Logger::get_logger();
my %ipv6_network_objects;

sub parse_system {
	my ($in, $out) = @_;

	# hostname is collected from its own command
	if( $in->{hostname} ) {
		$out->print_element('core:systemName', $in->{hostname});
	} else {
		$LOGGER->fatal("Could not find hostname for $in->{ip}, it is required for the backup to succeed");
	}

	# search through known check point servers for backup target
	my $server = get_server_by_ip($in, $in->{ip});

	# get config from hash repsonse
	my $version = $server->{version}            // 'N/A';
	my $os      = $server->{'operating-system'} // 'N/A';

	# print system information
	$out->open_element('core:osInfo');
	$out->print_element('core:make', 'Check Point');
	$out->print_element('core:version', $version);
	$out->print_element('core:osType', $os);

	# print packet processing order for Check Point gateways
	$out->open_element('core:packetProcessingOrdering');
	for my $phase ( 'packet-filtering', 'address-translation' ) {
		$out->open_element('core:packetProcessingElement');
		$out->print_element('core:packetProcessingPhase', $phase);
		$out->close_element('core:packetProcessingElement');
	}
	$out->close_element('core:packetProcessingOrdering');
	$out->close_element('core:osInfo');
	$out->print_element('core:deviceType', 'Firewall');

	return;
}

sub parse_chassis {
	my ($in, $out) = @_;

	# search through known check point servers for backup target
	my $server = get_server_by_ip($in, $in->{ip});

	# get config from hash repsonse
	my $model  = $server->{hardware} // 'N/A';
	my $serial = $server->{uid};

	# print chassis information
	my $chassis = {
		'core:asset' => {
			'core:assetType' => 'Chassis',
			'core:factoryinfo' => {
				'core:make' => 'Check Point',
				'core:modelNumber' => $model,
				'core:serialNumber' => $serial
			}
		}
	};
	$out->print_element('chassis', $chassis);

	return;
}

sub parse_interfaces {
	my ($in, $out) = @_;
	state %interfaces;

	if( not %interfaces ) {
		my $ip_regex = get_crep('cipm');

		while( $in->{ip_address} =~ /^\d+:\s+(\S+):\s+(.+?)(?=^\S|\Z)/msg ) {
			my ($name, $blob) = ($1, $2);
			my ($address, $mask) = $blob =~ /inet\s+($ip_regex)\/(\d+)/;
			my ($mac) = $blob =~ /link\/ether\s+(\S+)/; $mac =~ s/://g if $mac;
			my ($mtu) = $blob =~ /\s+mtu\s+(\d+)/;
			my $type = $blob =~ /loopback/i ? 'softwareLoopback' : 'ethernet';

			# vlan
			if( $name =~ /\d+\.\d+\@/ ) {
				($name) = $name =~ /(.+)\@/;
				$type = 'virtual';
			}

			my $status = $blob =~ /<\S*UP\S*>/ ? 'up' : 'down';
			my $physical = $type eq 'ethernet' ? 'true' : 'false';

			my $model = {
				name => $name,
				physical => $physical,
				adminStatus => $status,
				interfaceType => $type,
				mtu => $mtu,
				interfaceIp => {
					ipConfiguration => {
						ipAddress => $address,
						mask => $mask
					}
				}
			};

			$model->{interfaceEthernet} = { macAddress => $mac } if $mac;

			$interfaces{$name} = $model;
		}
	}

	if( $out ) {
		$out->open_element('interfaces');
		for my $interface ( sort keys %interfaces ) {
			$out->print_element('interface', $interfaces{$interface});
		}
		$out->close_element('interfaces');
	}

	return \%interfaces;
}

sub parse_object_groups {
	my ($in, $out) = @_;
	state %object_groups;

	if( not %object_groups ) {
		my $network_groups = _parse_network_groups($in);
		$object_groups{network_groups} = $network_groups;

		my $service_groups = _parse_service_groups($in);
		$object_groups{service_groups} = $service_groups;
	}

	if( $out ) {
		$out->open_element('objectGroups');
		my $network_groups = $object_groups{network_groups};
		my @elements = qw(objectGroupReference host range subnet);

		for my $network_group_name ( sort keys %{ $network_groups } ) {
			my $network_group = $network_groups->{$network_group_name};
			$out->open_element('networkGroup');
			$out->print_element('id', $network_group->{id});
			for my $element ( @elements ) {
				if(exists $network_group->{$element}){
					$out->print_element($element, $network_group->{$element});
				}
			}
			$out->close_element('networkGroup');
		}

		my $service_groups = $object_groups{service_groups};
		for my $service_group_name ( sort keys %{ $service_groups } ) {
			$out->print_element('serviceGroup', $service_groups->{$service_group_name});
		}

		$out->close_element('objectGroups');
	}

	return \%object_groups;
}

sub parse_filters {
	my ($in, $out) = @_;

	# get the active policy name
	my $server = get_server_by_ip($in, $in->{ip});
	my $policy_name = $server->{policy}->{'access-policy-name'};

	my $policy = get_policy_by_name($in, $policy_name);
	if( $policy ) {
		my $filter_lists = _parse_access_rulebase($in, $policy);
		my $nat_filter_lists = _parse_nat_rulebase($in, $policy_name);
		my $reverse_nat_filter_lists = _reverse_nat_rulebase($nat_filter_lists);
		my $interfaces = parse_interfaces($in);

		_print_filters($out, $interfaces, $filter_lists, $nat_filter_lists, $reverse_nat_filter_lists);
	}

	return;
}

sub parse_static_routing {
	my ($in, $out) = @_;
	my $ip_regex = get_crep('cipm');

	my @static_routes;
	while( $in->{route} =~ /^($ip_regex)\s+($ip_regex)\s+($ip_regex)\s+\S+\s+\d+\s+\d+\s+\d+\s+(\S+)\s*$/mg ) {
		my ($destination_ip, $gateway_ip, $destination_mask, $interface) = ($1, $2, $3, $4);
		my $default = $destination_ip eq '0.0.0.0' ? 'true' : 'false';
		my $model = {
			destinationAddress => $destination_ip,
			destinationMask => mask_to_bits($destination_mask),
			defaultGateway => $default,
			interface => $interface
		};

		# QRM cannot handle a gateway of 0.0.0.0, implies to use the default gateway anyway
		if( $gateway_ip ne '0.0.0.0' ) {
			$model->{gatewayAddress} = $gateway_ip;
		}
		push @static_routes, $model;
	}
	$out->print_element('staticRoutes', { staticRoute => \@static_routes } );

	return;
}


sub _parse_access_rulebase {
	my ($in, $policy) = @_;

	# Get the servre object of the device being backed up
	my $server = get_server_by_ip( $in, $in->{ip} );

	# objects groups are needed to validate objects the rules reference
	my $object_groups = parse_object_groups($in);
	my %network_groups = %{ $object_groups->{network_groups} };
	my %service_groups = %{ $object_groups->{service_groups} };

	# get firewall rulebases
	my $rulebase;
	for my $access_layer ( @{ $policy->{'access-layers'} } ) {
		if( $access_layer->{firewall} ) {
			$rulebase = $in->{'show-access-rulebase'}->($access_layer->{name});
			last; # got the access control rulebase
			#TODO 80.10 gateways will be able to have more than one access control layer
		}
	}
	my $rulebase_name = $rulebase->{name};
	$LOGGER->debug("Found access rulebase '$rulebase_name'");

	# rulebases are made of rules and rule sections, this gets one list of rules
	my $rules_or_sections = $rulebase->{rulebase};
	my @rules;
	for my $rule_or_section ( @{ $rules_or_sections } ) {
		if( $rule_or_section->{rulebase} ) {
			push @rules, @{ $rule_or_section->{rulebase} };
		} else {
			push @rules, $rule_or_section;
		}
	}
	my $rule_count = scalar @rules;
	$LOGGER->debug("Found $rule_count rules in '$rulebase_name'");

	# parse each rule to a filter entry model and store in an ordered list
	# the first element of which is the name of the acl
	my @filters = ( $rulebase_name );

	RULE:
	for my $rule ( @rules )  {
		next unless $rule->{enabled};

		my $name = $rule->{name};

		my $action = lc $rule->{action}->{name} eq 'accept' ? 'permit' : 'deny' ;
		my $model = {
			log => 'true',
			name => $name,
			primaryAction => $action,
			processOrder => $rule->{'rule-number'},
			ztSignature => $rule->{uid}
		};

		# parse rule's IPv4 addresses
		my $dropped_ipv6 = 0;
		for my $direction ( qw(source destination) ) {
			if( $rule->{"$direction-negate"} ) {
				$LOGGER->debug("WARNING: Dropping rule '$name', $direction-negate is not supported");
				next RULE;
			}

			my @sub_models;
			my $objects = $rule->{$direction};
			for my $object ( @{ $objects } ) {
				my $id = $object->{name};
				if( $network_groups{$id} ) {
					my $sub_model = { objectGroupReference => $id };
					push @sub_models, $sub_model;
				} elsif( $id ne 'Any' ) {
					if( $ipv6_network_objects{$id} ) {
						$dropped_ipv6 = 1;
					} else {
						$LOGGER->debug("WARNING: Dropping rule '$name', $direction address '$id' is unknown or invalid");
						next RULE;
					}
				}
			}
			$model->{$direction . 'IpAddr'} = \@sub_models if @sub_models;
		}

		if( $dropped_ipv6 ) {
			unless( $model->{sourceIpAddr} or $model->{destinationIpAddr} ) {
				$LOGGER->debug("WARNING: Dropping rule '$name', it contains only IPv6 address information");
				next RULE;
			}
		}

		# parse rule's services
		if( $rule->{'service-negate'} ) {
			$LOGGER->debug("WARNING: Dropping rule '$name', service-negate is not supported");
			next RULE;
		}

		my @source_services;
		my @destination_services;
		my $services = $rule->{service};
		for my $service ( @{ $services } ) {
			my $service_id = $service->{name};
			my $sub_model;
			if( $service_groups{$service_id} ) {
				$sub_model = { objectGroupReference => $service_id };
				push @destination_services, $sub_model;
			}
			if( $service_groups{$service_id . '__src'} ) {
				$sub_model = { objectGroupReference => $service_id . '__src' };
				push @source_services, $sub_model;
			}
			if( $service_id ne 'Any' and not $sub_model ) {
				$LOGGER->debug("WARNING: Dropping rule '$name', services '$service_id' and '$service_id\__src' are unknown or invalid");
				next RULE;
			}
		}
		$model->{sourceService} = \@source_services if @source_services;
		$model->{destinationService} = \@destination_services if @destination_services;

		my $install = 0;
		for my $target ( @{ $rule->{'install-on'} } ) {
			my $target_name = $target->{name};
			if( $target_name eq 'Policy Targets' or $target_name eq $server->{name} ) {
				$install = 1; last;
			}
		}
		if( not $install ) {
			$LOGGER->debug("WARNING: Dropping rule '$name', '$server->{name}' is not an install target of this rule");
			next RULE;
		}

		push @filters, $model;
	}

	return \@filters;
}

sub _parse_nat_rulebase {
	my ($in, $policy_name) = @_;

	# Get the servre object of the device being backed up
	my $server = get_server_by_ip( $in, $in->{ip} );

	# objects groups are needed to validate objects the rules reference
	my $object_groups = parse_object_groups($in);
	my %network_groups = %{ $object_groups->{network_groups} };
	my %service_groups = %{ $object_groups->{service_groups} };

	# get firewall rulebases
	my $rulebase = $in->{'show-nat-rulebase'}->($policy_name);
	my $rulebase_name = "$policy_name NAT";

	# rulebases are made of rules and rule sections, this gets one list of rules
	my $rules_or_sections = $rulebase->{rulebase};
	my @rules;
	for my $rule_or_section ( @{ $rules_or_sections } ) {
		if( $rule_or_section->{rulebase} ) {
			push @rules, @{ $rule_or_section->{rulebase} };
		} else {
			push @rules, $rule_or_section;
		}
	}
	my $rule_count = scalar @rules;
	$LOGGER->debug("Found $rule_count rules in '$rulebase_name'");

	# parse each rule to a filter entry model and store in an ordered list
	# the first element of which is the name of the nat acl
	my @filters = ( $rulebase_name );

	RULE:
	for my $rule ( @rules )  {
		next unless $rule->{enabled};

		my $name = $rule->{uid};
		my $model = {
			log => 'true',
			primaryAction => 'none',
			processOrder => $rule->{'rule-number'},
			ztSignature => $rule->{uid}
		};

		# parse rule's original IPv4 addresses
		for my $direction ( qw(source destination) ) {
			my $object = $rule->{"original-$direction"};
			my $id = $object->{name};
			if ($id eq "Any"){
				next;
			}elsif (!$network_groups{$id}){
				$LOGGER->debug("WARNING: Dropping NAT rule '$name', $direction address '$id' is unknown or invalid");
				next RULE;
			}
			my $isAutoRule = $object->{'nat-settings'}->{'auto-rule'} // "false";
			if ($isAutoRule eq "true" && $direction eq "destination"){
				parse_auto_nat_settings ($direction, $object, $model, $in->{ip});
			}else{
				$model->{$direction . 'IpAddr'} = { objectGroupReference => $id };
			}
		}

		for my $direction ( qw(source destination) ) {
			my $object = $rule->{"translated-$direction"};
			my $id = $object->{name};
			if ($id eq "Original"){
				next; # skip object if no translation
			}elsif (!$network_groups{$id}){
				$LOGGER->debug("WARNING: Dropping NAT rule '$name', $direction translation address '$id' is unknown or invalid");
				next RULE;
			}
			my $isAutoRule = $object->{'nat-settings'}->{'auto-rule'} // "false";
			if ($isAutoRule eq "true" && $direction eq "source"){
				parse_auto_nat_settings ($direction, $object, $model, $in->{ip});
			} else{	
				my $model_direction = $direction eq 'source' ? $direction : 'dest';
				$model->{$model_direction . 'AddressTranslation'} = {
					translateAppliedAddressing => 'true',
					translationAddress => {
						translationIP => {
						objectGroupReference => $id
						}
					}
				};
			}
		}

		my $service = $rule->{'original-service'};
		my $service_id = $service->{name};
		my $sub_model;
		if( $service_groups{$service_id} ) {
			$sub_model = { objectGroupReference => $service_id };
			$model->{destinationService} = $sub_model;
		}
		if( $service_groups{$service_id . '__src'} ) {
			$sub_model = { objectGroupReference => $service_id . '__src' };
			$model->{sourceService} = $sub_model;
		}
		if( $service_id ne 'Any' and not $sub_model ) {
			$LOGGER->debug("WARNING: Dropping NAT rule '$name', services '$service_id' and '$service_id\__src' are unknown or invalid");
			next RULE;
		}

		my $translated_service = $rule->{'translated-service'};
		my $translated_service_id = $translated_service->{name};
		if( $translated_service_id ne 'Original' ) {
			my $protocol = get_protocol_for_service(\%service_groups, $translated_service_id);
			my $translated_sub_model;
			if( $service_groups{$translated_service_id} ) {
				$translated_sub_model = { objectGroupReference => $translated_service_id };
				$model->{destAddressTranslation}->{translationAddress}->{translationServicePort} = {
					translationPort => $translated_sub_model,
					translationProtocol => $protocol
				};
			}
			if( $service_groups{$translated_service_id . '__src'} ) {
				$translated_sub_model = { objectGroupReference => $translated_service_id . '__src' };
				$model->{sourceAddressTranslation}->{translationAddress}->{translationServicePort} = {
					translationPort => $translated_sub_model,
					translationProtocol => $protocol
				};
			}
			if( not $translated_sub_model ) {
				$LOGGER->debug("WARNING: Dropping NAT rule '$name', translation services '$translated_service_id' and '$translated_service_id\__src' are unknown or invalid");
				next RULE;
			}
		}

		for my $target ( @{ $rule->{'install-on'} } ) {
			my $target_name = $target->{name};
			if( $target_name ne 'Policy Targets' and $target_name ne 'All' and $target_name ne $server->{name} ) {
				$LOGGER->debug("INFO: Dropping NAT rule '$name', '$server->{name}' is not an install target of this rule");
				next RULE;
			}
		}

		push @filters, $model;
	}

	return \@filters;
}

sub _reverse_nat_rulebase {
	my ($nat_filters) = @_;
	my @nat_filters = @{ $nat_filters };
	my @reverse_filters;

	my $list_name = shift @nat_filters;
	for my $rule ( @nat_filters ) {
		if ($rule->{isAutoRule}){
			delete $rule->{isAutoRule};
			next; # do not reverse automatic rules as already included in rulebase
		}
		my $reverse_source_translation;
		if( $rule->{sourceIpAddr} ) {
			$reverse_source_translation = {
				translateAppliedAddressing => 'true',
				translationAddress => {
					translationIP => $rule->{sourceIpAddr}
				}
			};
		}

		my $reverse_destination_translation;
		if( $rule->{destinationIpAddr} ) {
			$reverse_destination_translation = {
				translateAppliedAddressing => 'true',
				translationAddress => {
					translationIP => $rule->{destinationIpAddr}
				}
			};
		}

		my $reverse_source;
		my $reverse_source_service;
		my $reverse_source_protocol;
		if( $rule->{sourceAddressTranslation} ) {
			my $translation_address = $rule->{sourceAddressTranslation}->{translationAddress};
			$reverse_source = $translation_address->{translationIP};
			if( $translation_address->{translationServicePort} ) {
				$reverse_source_service = $translation_address->{translationServicePort}->{translationPort};
				$reverse_source_protocol = $translation_address->{translationServicePort}->{translationProtocol};
			}
		}

		my $reverse_destination;
		my $reverse_destination_service;
		my $reverse_destination_protocol;
		if( $rule->{destAddressTranslation} ) {
			my $translation_address = $rule->{destAddressTranslation}->{translationAddress};
			$reverse_destination = $translation_address->{translationIP};
			if( $translation_address->{translationServicePort} ) {
				$reverse_destination_service = $translation_address->{translationServicePort}->{translationPort};
				$reverse_destination_protocol = $translation_address->{translationServicePort}->{translationProtocol};
			}
		}

		my $reverse_source_service_translation;
		if( $rule->{sourceService} ) {
			$reverse_source_service_translation = {
				translationPort => $rule->{sourceService},
				translationProtocol => $reverse_source_protocol
			};
		}

		my $reverse_destination_service_translation;
		if( $rule->{destinationService} ) {
			$reverse_destination_service_translation = {
				translationPort => $rule->{destinationService},
				translationProtocol => $reverse_destination_protocol
			};
		}

		my $reverse_rule = dclone $rule;
		for my $element ( qw( sourceIpAddr destinationIpAddr sourceAddressTranslation destAddressTranslation sourceService destinationService ) ) {
			delete $reverse_rule->{$element};
		}
		$reverse_rule->{destinationIpAddr} = $reverse_source if $reverse_source;
		$reverse_rule->{sourceIpAddr} = $reverse_destination if $reverse_destination;
		$reverse_rule->{destAddressTranslation} = $reverse_source_translation if $reverse_source_translation;
		$reverse_rule->{sourceAddressTranslation} = $reverse_destination_translation if $reverse_destination_translation;
		$reverse_rule->{destAddressTranslation}->{translationAddress}->{translationServicePort} = $reverse_source_service_translation if $reverse_source_service_translation;
		$reverse_rule->{sourceAddressTranslation}->{translationAddress}->{translationServicePort} = $reverse_destination_service_translation if $reverse_destination_service_translation;
		$reverse_rule->{destinationService} = $reverse_source_service if $reverse_source_service;
		$reverse_rule->{sourceService} = $reverse_destination_service if $reverse_destination_service;

		# Add the rule, unless it is going to match all traffic
		if( $reverse_rule->{destinationIpAddr} or $reverse_rule->{sourceIpAddr} ) {
			push @reverse_filters, $reverse_rule;
		}
	}

	# When printing lists the first element is expected to be a string, the list name
	unshift @nat_filters, $list_name;
	unshift @reverse_filters, "$list_name Implied Reverse";

	return \@reverse_filters;
}

sub _print_filters {
	my ($out, $interfaces, $access_filters, $nat_filters, $reverse_nat_filters) = @_;
	my @phases = ( 'packet-filtering', 'address-translation', 'address-translation' );
	my @phase_positions = ( '', 'routing:before', 'routing:after');
	my @default_actions = ( 'deny', 'none', 'none' );

	# interface names are needed for the applies to elements
	my @interface_names;
	for my $interface_name ( sort keys %{ $interfaces } ) {
		if( $interfaces->{$interface_name}->{interfaceType} ne 'softwareLoopback' ) {
			push @interface_names, $interface_name;
		}
	}

	# print the parsed filter lists, some elements are not alphabeticaly ordered
	$out->open_element('filterLists');
	for my $list ( $access_filters, $nat_filters, $reverse_nat_filters ) {
		my @filters = @{ $list };
		my $list_name = shift @filters;

		$out->open_element('filterListsSection');
		$out->open_element('filterList');

		my $default_action = shift @default_actions;
		$out->print_element('defaultAction', $default_action);

		for my $source_interface ( @interface_names ) {
			for my $destination_interface ( @interface_names ) {
				$out->open_element('appliesTo');
				$out->print_element('source', $source_interface);
				$out->print_element('destination', $destination_interface);
				$out->close_element('appliesTo');
			}
		}
		$out->print_element('filterEntry', \@filters );
		$out->print_element('mode', 'stateful');
		$out->print_element('name', $list_name);

		my $phase = shift @phases;
		$out->print_element('processingPhase', { 'core:packetProcessingPhase' => $phase } );

		$out->close_element('filterList');

		my $phase_position = shift @phase_positions;
		if( $phase_position ) {
			($phase, my $position) = split /:/, $phase_position;
			$out->print_element('phaseDiscriminator', { processingPhasePosition => $position, processingPhase => { 'core:packetProcessingPhase' => $phase } } );
		}
		$out->close_element('filterListsSection');
	}
	$out->close_element('filterLists');

	return;
}

sub _parse_network_groups {
	my ($in) = @_;
	my %network_groups;

	my $hosts = _parse_hosts($in);
	my $networks = _parse_networks($in);
	my $ranges = _parse_ranges($in);
	%network_groups = ( %{ $hosts }, %{ $networks }, %{ $ranges } );

	my $groups = _parse_address_groups($in, \%network_groups);
	%network_groups = ( %network_groups, %{ $groups } );

	my $group_with_exclusions = _parse_address_group_with_exclusions($in, \%network_groups);
	%network_groups = ( %network_groups, %{ $group_with_exclusions } );

	return \%network_groups;
}

sub _parse_hosts {
	my ($in) = @_;
	my %network_groups;

	my $simple_gateways = $in->{'show-simple-gateways'};
	my $hosts = $in->{'show-hosts'};
	for my $response ( $simple_gateways, $hosts ) {
		for my $host ( @{ $response } ) {
			my $id = $host->{name};
			my $ip = $host->{'ipv4-address'};

			if( $ip ) {
				my $model = {
					id => $id,
					host => [ $ip ]
				};
				$network_groups{$id} = $model;
			} else {
				$LOGGER->debug("WARNING: Dropping network object '$id', no IPv4 address found");
				$ipv6_network_objects{$id} = 1;
			}
		}
	}

	return \%network_groups;
}

sub _parse_networks {
	my ($in) = @_;
	my %network_groups;

	my $networks = $in->{'show-networks'};
	for my $network ( @{ $networks } ) {
		my $id = $network->{name};
		my $address = $network->{subnet4};
		my $mask = $network->{'mask-length4'};
		if( $address ) {
			my $model = {
				id => $id,
				subnet => [ {
					address => $address,
					mask => $mask
				} ]
			};
			$network_groups{$id} = $model;
		} else {
			$LOGGER->debug("WARNING: Dropping network object '$id', no IPv4 network found");
			$ipv6_network_objects{$id} = 1;
		}
	}

	return \%network_groups;
}

sub _parse_ranges {
	my ($in) = @_;
	my %network_groups;

	my $ranges = $in->{'show-address-ranges'};
	for my $range ( @{ $ranges } ) {
		my $id = $range->{name};
		my $start = $range->{'ipv4-address-first'};
		my $end = $range->{'ipv4-address-last'};

		if( $start ) {
			my $model = {
				id => $id,
				range => [ {
					startAddress => $start,
					endAddress => $end
				} ]
			};
			$network_groups{$id} = $model;
		} else {
			$LOGGER->debug("WARNING: Dropping network object '$id', no IPv4 range found");
			$ipv6_network_objects{$id} = 1;
		}
	}

	return \%network_groups;
}

sub _parse_address_groups {
	my ($in, $network_groups) = @_;
	my %network_groups = %{ $network_groups };

	my $groups = $in->{'show-groups'};
	for my $group ( @{ $groups } ) {
		my $id = $group->{name};
		my $members = $group->{members};
		my $model = {
			id => $id
		};

		my @member_ids;
		my $ipv6_member = 0;
		for my $member ( @{ $members } ) {
			my $member_id = $member->{name};
			if( not $network_groups{$member_id} ) {
				$LOGGER->debug("WARNING: Dropping network object member from '$id', unknown or invalid member '$member_id'");
				if( $ipv6_network_objects{$member_id} ) {
					$ipv6_member = 1;
				}
			} else {
				push @member_ids, $member_id;
			}
		}

		if( @member_ids ) {
			$model->{objectGroupReference} = \@member_ids;
			$network_groups{$id} = $model;
		} else {
			$LOGGER->debug("WARNING: Dropping network object '$id', no valid members found");
			if( $ipv6_member ) {
				$ipv6_network_objects{$id} = 1;
			}
		}
	}

	return \%network_groups;
}

sub _parse_address_group_with_exclusions {
	my ($in, $network_groups) = @_;
	my %network_groups = %{ $network_groups };

	my $groups_with_exclusion = $in->{'show-groups-with-exclusion'};
	for my $group_with_exclusion ( @{ $groups_with_exclusion } ) {
		my $id = $group_with_exclusion->{name};
		my $include = $group_with_exclusion->{include}->{name};
		my $except = $group_with_exclusion->{except}->{name};
		my $include_set; 
		if( $include =~ /any/i ) {
			$include_set = Net::CIDR::Set->new('0.0.0.0/0');
		} else {
			$include_set = get_network_group_set($include, \%network_groups);
		}
		my $except_set = get_network_group_set($except, \%network_groups);
		if( $except_set ) {
			my $diff = $include_set->diff($except_set);
			my @diffs = $diff->as_range_array();
			my @ranges;
			my @subnets;
			for my $diff( @diffs ) {
				if( $diff =~ /(.+)-(.+)/ ) {
					push @ranges, { startAddress => $1, endAddress => $2 };
				} else {
					my ($address, $mask) = split /\//, $diff;
					push @subnets, { address => $address, mask => $mask // 32 };
				}
			}

			my $model = { id => $id };
			$model->{range} = \@ranges if @ranges;
			$model->{subnet} = \@subnets if @subnets;
			$network_groups{$id} = $model;
		}
	}

	return \%network_groups;
}

sub _parse_service_groups {
	my ($in) = @_;
	my %service_groups;

	# parse tcp and udp services, creating a source service if there is a source port
	for my $protocol ( qw(tcp udp) ) {
		my $services = $in->{"show-services-$protocol"};

		for my $service ( @{ $services } ) {
			my $id = $service->{name};
			my $port = $service->{port};
			$service_groups{$id} = _parse_service($id, $port, $protocol);

			# check for source port and create a service group for it
			my $source_port = $service->{'source-port'};
			if( defined $source_port ) {
				my $source_id = $id . '__src';
				$service_groups{$source_id} =_parse_service($source_id, $source_port, $protocol);
			}
		}
	}

	# parse groups of services
	my $group_services = $in->{'show-service-groups'};
	for my $group ( @{ $group_services } ) {
		my $id = $group->{name};
		my $members = $group->{members};

		my $model = { id => $id };
		my @member_ids;
		for my $member ( @{ $members } ) {
			my $member_id = $member->{name};
			if( $service_groups{$member_id} ) {
				push @member_ids, $member_id;
			} else {
				$LOGGER->debug("WARNING: Dropping service object member from '$id', unknown or invalid member '$member_id'");
			}
		}

		if( @member_ids ) {
			$model->{objectGroupReference} = \@member_ids;
			$service_groups{$id} = $model;
		} else {
			$LOGGER->debug("WARNING: Dropping service object '$id', no valid members found");
		}
	}

	return \%service_groups;
}

sub _parse_service {
	my ($id, $port, $protocol) = @_;
	my $model = { id => $id };
	if( $port =~ /^([<>])?(\d+)$/ ) {
		my $operator = $1 ? $1 eq '>' ? 'gt' : 'lt' : 'eq';
		$model->{portExpression} = [ { operator => $operator, port => $2, protocol => $protocol } ];
	} elsif( $port =~ /^(\d+)-(\d+)$/ ) {
		$model->{portRange} = [ { portStart => $1, portEnd => $2, protocol => $protocol } ];
	}
	return $model;
}

1;

sub parse_auto_nat_settings {
	my ($direction, $object, $model, $gatewayIP) = @_;
	
	$model->{isAutoRule} = 1;
	my $natSettings = $object->{'nat-settings'};
	my $NATtype =  $natSettings->{'method'};
	if ($NATtype eq "static"){
		parse_auto_static_nat ($natSettings, $model, $direction);
	}elsif ($NATtype eq "hide"){
		parse_auto_hide_nat ($direction, $object, $natSettings, $gatewayIP, $model);
	}
	return;
}

sub parse_auto_static_nat {
	my ($natSettings, $model, $direction, $isHideNAT) = @_;

	my $ip = $natSettings->{'ipv4-address'};
	return unless $ip; # ignore translation if no IPv4 address
	if ($direction eq "source"){
		$model->{sourceAddressTranslation} = {
			translateAppliedAddressing => 'true',
			translationAddress => {
					translationIP => {
						host => $ip
					}
			}
		};
	} else {
		$model->{'destinationIpAddr'} = { host => $ip };
	}
	return;
}

sub parse_auto_hide_nat {
	my ($direction, $object, $natSettings, $gatewayIP, $model) = @_;

	if ( $direction eq "destination" ) {
		my $id = $object->{name};
	    $model->{$direction . 'IpAddr'} = { objectGroupReference => $id };
	}else{
		my $hideBehind = $natSettings->{'hide-behind'};
		if ($hideBehind eq "gateway"){
			$model->{'sourceAddressTranslation'} = {
				translateAppliedAddressing => 'true',
				translationAddress => {
						translationIP => {
							host => $gatewayIP # gateway IP
						},
				}
			};
		}elsif ($hideBehind eq "ip-address"){
			# same translation as static nat but with PAT
			parse_auto_static_nat($natSettings, $model, $direction, 1);
		}
		# add PAT for default ranges 600-1023 and 10000-60000
		$model->{'sourceAddressTranslation'}->{translationAddress}->{translationServicePort} = [
			{translationPort => {portRange =>{portStart => 600 , portEnd => 1023}}, translationProtocol => 'tcp'},
			{translationPort => {portRange =>{portStart => 10000 , portEnd => 60000}}, translationProtocol => 'tcp'}
		]
	}
	return;
}

1;